// IFF_ILBM_Handler

#include "DataHandler.h"
#include "Datatypes.h"
#include "DataStreams.h"
#include <InterfaceDefs.h>
#include <Debug.h>


char handlerName[] = "Amiga IFF/ILBM";
long handlerVersion = 9;
char handlerInfo[] = "Demo version .9";


#define DUMP_FORMAT		'BITM'

Format inputFormats[] = {
	{	DATA_BITMAP,	DATA_BITMAP,
		0.9,	0.5,
		"",
		"Amiga IFF/ILBM"
	},
	{	0,	0,
		0,	0,
		"",
		""
	}
};

Format outputFormats[] = {
	{	DATA_BITMAP,	DATA_BITMAP,
		0.9,	0.5,
		"",
		"Amiga IFF/ILBM"
	},
	{	0,	0,
		0,	0,
		"",
		""
	}
};

/* we use the same parser for Identify and Translate */
enum {
	MODE_IDENTIFY,
	MODE_TRANSLATE
};

bool process_chunks( DStream &inStream, DStream *outStream, int mode );

/* Calculate the number of bytes/row, unfortunately the original
   function from dumphandler.c was broken when dealing with
   B_COLOR_8_BIT.  Each row needs to be long aligned */
static int
GetBITMRowBytes(
	long			left,
	long			right,
	color_space		colors)
{
	int rowBytes = right-left+1;	//	calculate a rowBytes to work with
	switch (colors) {
	case B_COLOR_8_BIT:
		rowBytes = ((rowBytes + 3) / 4) * 4;
		break;
	case B_RGB_32_BIT:
		rowBytes = rowBytes*4; break;
	case B_RGB_16_BIT:
		rowBytes = rowBytes*2; break;
	case B_MONOCHROME_1_BIT:
		rowBytes = rowBytes/8; break;
	case B_GRAYSCALE_8_BIT:
		/* rowBytes are OK */ break;
	default:
		/* illegal format - no rowBytes */
		rowBytes = 0; break;
	}
	return rowBytes;
}

/* Identify function */
long
Identify(
	DStream &			inSpec,
	const Format *		format,
	BMessage *			/* ioExtension */,
	DATAInfo &			outInfo,
	ulong				outType)
{
	bool isDataBitmap = false;
	
	/* if the hint type is not DATA_BITMAP we need to wretch */
	if (outType && (outType != DATA_BITMAP))
		return DATA_NO_HANDLER;

	if( (format) && (format->type != inputFormats[0].type) )
		return DATA_NO_HANDLER;

	/* parse the file */
	isDataBitmap = process_chunks( inSpec, 0, MODE_IDENTIFY );
	if( isDataBitmap )
	{
		/* it is a bitmap set the DATAInfo passed in */
		format = &inputFormats[0];
		outInfo.formatType = format->type;
		outInfo.formatGroup = format->t_cls;
		outInfo.formatQuality = format->quality;
		outInfo.formatCapability = format->capability;
		strcpy(outInfo.formatName, format->formatName);
		strcpy(outInfo.MIMEName, format->MIME);
		return B_OK;
	}
	return( DATA_NO_HANDLER );
}

/* Amiga IFF stuff */
typedef unsigned short UWORD;
typedef short WORD;
typedef unsigned char UBYTE;
typedef char BYTE;

typedef UBYTE Masking;

#define mskNone 0
#define mskHasMask 1

typedef UBYTE Compression;

#define cmpNone 0
#define cmpByteRun1 1

#define DATA_SIZE 12

struct BitmapHeader {
	UWORD w, h;
	WORD x, y;
	UBYTE nPlanes;
	Masking masking;
	Compression compression;
	UBYTE pad1;
	UWORD transparentColor;
	UBYTE xAspect, yAspect;
	WORD pageWidth, pageHeight;
};

#define HAM 0x800
#define EXTRA_HALFBRITE 0x80

struct ColorRegister {
	char red;
	char green;
	char blue;
};

/* Parses the file for us */
bool process_chunks( DStream &inStream, DStream *outStream, int mode )
{
	unsigned long curr_chunk;
	long e = 0;
	bool dead = false;
	long name_size = 4, chunk_size, size;
	bool isDataBitmap = false, found_bmh = false;
	unsigned char *cmap = 0;
	UBYTE *real_cmap = 0;
	char *body = 0;
	char *row = 0;
	DATABitmap hdr;
	struct BitmapHeader bmh;
	char	data[DATA_SIZE];
	unsigned long modeid = 0;
	bool ham = false, ehb = false;

	/* how much initial stuff do we need.  FORM(size)ILBM so its 12 */
	size = DATA_SIZE;
	e = inStream.Read( data, size );
	if( !e && (size < DATA_SIZE) )
		e = DATA_NO_HANDLER;
	if( e < 0 )
		return e;

	/* check if its a real IFF file */
	if( strncmp( data, "FORM", 4 ) == 0 )
	{
		unsigned long length;
		
		length = *((long *)&data[4]);
#ifdef DEBUG_IFF
		printf( "found IFF of size %ld\n", length );
#endif
		/* its IFF so we need to check if its ILBM */
		if( strncmp( &data[8], "ILBM", 4 ) == 0 )
		{
			isDataBitmap = true;
			length -= 4;
			hdr.magic = DATA_BITMAP;
			/* we loop so long as there is something to read (length), there
			   hasn't been some error (!dead), and there wasn't an error in
			   reading (inStream.Read() */
			while( length && !dead && ((e = inStream.Read( &curr_chunk, name_size )) == 0) )
			{
				size = 4;
				/* we've read the chunk name now we need to get the
				   size of the chunk */
				if( (e = inStream.Read( &chunk_size, size )) == 0 )
				{
					length -= 8;
					length -= chunk_size;
					switch( curr_chunk )
					{
						case 'BMHD':
							/* we get the BitmapHeader struct */
#ifdef DEBUG_IFF
							printf( "BMHD %ld\n", chunk_size );
#endif
							size = sizeof( struct BitmapHeader );
							if( (e = inStream.Read( &bmh, size )) == 0 )
							{
								found_bmh = true;
								hdr.bounds.Set( 0, 0, bmh.w - 1, bmh.h - 1 );
								/* find the depth */
								if( bmh.nPlanes == 24 )
								{
									hdr.colors = B_RGB_32_BIT;
#ifdef DEBUG_IFF
									printf( "32bit\n" );
#endif
								}
								else if( (bmh.nPlanes > 0) && (bmh.nPlanes <= 8) )
								{
									hdr.colors = B_COLOR_8_BIT;
#ifdef DEBUG_IFF
									printf( "8bit\n" );
#endif
								}
								else
									isDataBitmap = false;
								if( bmh.masking == mskHasMask )
									bmh.nPlanes++;
								else
									bmh.masking = mskNone;
								/* using the depth we need to get the number of row bytes and
								   compute the total size of the data */
								hdr.rowBytes = GetBITMRowBytes( hdr.bounds.left, hdr.bounds.right, hdr.colors );
								hdr.dataSize = ((long)hdr.bounds.Height()+1)*hdr.rowBytes;
#ifdef DEBUG_IFF
								printf( "x: %ld  y: %ld  w: %ld  h: %ld\n", bmh.x, bmh.y, bmh.w, bmh.h );
								printf( "rb: %ld  ds: %ld\n", hdr.rowBytes, hdr.dataSize );
								printf( "msk: %ld  cmp: %ld  depth: %ld\n", bmh.masking, bmh.compression, bmh.nPlanes );
#endif
								/* we don't have to go through the rest of the file
								   on an identify */
								if( mode == MODE_IDENTIFY )
									length = 0;
							}
							else
								dead = true;
							break;
						case 'CMAP':
							/* colormap */
#ifdef DEBUG_IFF
							printf( "CMAP %ld\n", chunk_size );
#endif
							if( mode == MODE_IDENTIFY )
							{
								inStream.Seek( inStream.Position() + chunk_size );
							}
							else
							{
								int lpc;
								unsigned char red, green, blue;
								unsigned char *place;
								
								/* read in the colormap */
								cmap = (unsigned char *)malloc( chunk_size );
								real_cmap = (unsigned char *)malloc( 256 );
								if( cmap && real_cmap && ((e = inStream.Read( cmap, chunk_size)) == 0) )
								{
									/* convert the colormap to 8 bit if it already isn't and
									   then find the indexes to the standard colormap */
									place = cmap;
									for( lpc = 0; lpc < (chunk_size / 3); lpc++ )
									{
#ifdef DEBUG_IFF
										printf( "0x%x ", *place );
#endif
										/* if its 0xf0 we assume its really just a 4 bit color value
										   so we duplicate the value in the low order bits so now its
										   0xff, that way white is 0xff instead of 0xf0 */
										if( !(*place & 0x0f) )
											*place = (*place) | (*place >> 4);
										red = *place;
										place++;
#ifdef DEBUG_IFF
										printf( "0x%x ", *place );
#endif
										if( !(*place & 0x0f) )
												*place = (*place) | (*place >> 4);
										green = *place;
										place++;
#ifdef DEBUG_IFF
										printf( "0x%x\n", *place );
#endif
										if( !(*place & 0x0f) )
													*place = (*place) | (*place >> 4);
										blue = *place;
										place++;
#ifdef DEBUG_IFF
										printf( "  %ld  %lx %lx %lx\n", lpc, red, green, blue );
#endif
										/* find the index to the system map and put it in the
										   second map */
										real_cmap[lpc] = index_for_color( red, green, blue );
									}
									/* if there are 32 colors and 6 planes then we assume its ehb */
									if( ((chunk_size / 3) == 32) && (bmh.nPlanes == 6) )
									{
										place = cmap;
										for( lpc = 32; lpc < 64; lpc++ )
										{
											/* fill in the higher order table elements with
											half brite colors */
											red = *place / 2;
											green = *place / 2;
											blue = *place / 2;
											real_cmap[lpc] = index_for_color( red, green, blue );
										}
									}
								}
								else
									dead = true;
							}
							break;
						case 'GRAB':
							inStream.Seek( inStream.Position() + chunk_size );
							break;
						case 'DEST':
							inStream.Seek( inStream.Position() + chunk_size );
							break;
						case 'SPRT':
							inStream.Seek( inStream.Position() + chunk_size );
							break;
						case 'CAMG':
							/* modeid chunk */
#ifdef DEBUG_IFF
							printf( "CAMG %ld\n", chunk_size );
#endif
							size = 4;
							if( (e = inStream.Read( &modeid, size )) == 0 )
							{
#ifdef DEBUG_IFF
								printf( "modeid: %lx\n", modeid );
#endif
								/* if its ham we need to change to 32 bit and recompute
								   rowBytes and dataSize */
								if( modeid & HAM )
								{
									ham = true;
									hdr.colors = B_RGB_32_BIT;
									hdr.rowBytes = GetBITMRowBytes( hdr.bounds.left, hdr.bounds.right, hdr.colors );
									hdr.dataSize = ((long)hdr.bounds.Height()+1)*hdr.rowBytes;
								}
								if( modeid & EXTRA_HALFBRITE )
								{
									ehb = true;
								}
							}
							else
								dead = true;
							break;
						case 'CRNG':
							inStream.Seek( inStream.Position() + chunk_size );
							break;
						case 'CCRT':
							inStream.Seek( inStream.Position() + chunk_size );
							break;
						case 'BODY':
							/* the planes */
							if( mode == MODE_IDENTIFY )
							{
								//inStream.Seek( inStream.Position() + chunk_size );
								isDataBitmap = false;
								dead = true;
							}
							else
							{
								/* functions? we don't need no stinking functions! */
								int curr_row, curr_pix, curr_plane;
								int lpc;
								int rowWords = ((bmh.w + 15)/16) * 16;
								char *place, *row24 = 0;
								char plane_bit;
								int init, skip;
								
								/* the default format for storing 24 bit data in IFF
								   is a little brain dead.  There are 24 planes where
								   the first plane is r7 then r6, r5, r4, and so on.
								   Then it will go onto green and blue.  So when we
								   put it into the 32 bit row data we have to skip
								   over other color bytes. */
								if( (hdr.colors == B_RGB_32_BIT) && !ham )
								{
									/* this is 2 since BBitmap::Bits() has the colors
									   ordered blue green red alpha, instead of red
									   green blue */
									init = 2;
									skip = 4;
									rowWords *= 4;
								}
								else
								{
									init = 0;
									skip = 1;
								}
								row = (char *)malloc( rowWords ); /* a row of pixels to be written out */
								body = (char *)malloc( chunk_size ); /* the whole body data read in */
								place = body;
								if( ham )
								{
									/* if its ham we have to read the planes in regularly and then
									   convert them to a 32bit row which we will write out */
									row24 = (char *)malloc( rowWords * 4 );
									if( row24 )
									{
										/* if the first pixel in a row is a modify then it
										   modifies the border pixel which is the 0 color
										   register so we need to set this up instead of
										   writing a special case in the ham -> 32 bit code */
										row24[0] = cmap[2];
										row24[1] = cmap[1];
										row24[2] = cmap[0];
										row24[3] = 0;
										row24 += 4;
									}
									else
										dead = true;
								}
								if( body && row && !dead &&
									(cmap || (hdr.colors == B_RGB_32_BIT)) &&
									((e = inStream.Read( body, chunk_size )) == 0) )
								{
									outStream->Write( &hdr, sizeof( DATABitmap ) );
#ifdef DEBUG_IFF
									printf( "rw: %ld  rb: %ld\n", rowWords, hdr.rowBytes );
#endif
									/* loop through every row in the picture */
									for( curr_row = 0; (curr_row < bmh.h) && !dead; curr_row++ )
									{
#ifdef DEBUG_IFF
										printf( "row: %ld\n", curr_row );
#endif
										/* clean out the last row we read in */
										for( lpc = 0; lpc < rowWords; lpc++ )
										{
											row[lpc] = 0;
										}
										/* planes are interleaved which is groovy so we can read in the planes
										   convert to a chunky row and then write them out */
										for( curr_plane = 0; (curr_plane < bmh.nPlanes) && !dead; curr_plane++ )
										{
											/* if its 32 bit we need to reset the init value */
											if( (hdr.colors == B_RGB_32_BIT) && !ham )
											{
												if( curr_plane == 0 )
													init = 2;
												else if( !(curr_plane % 8) )
														init--;
											}
#ifdef DEBUG_IFF
											printf( "  plane: %ld  init: %ld\n", curr_plane, init );
#endif
											/* current bit in the chunky pixel we're reading in */
											plane_bit = (1 << (curr_plane % 8));
											/* check for compression */
											if( bmh.compression == cmpNone )
											{
												char *curr_place, *row_place;
												
												for( curr_place = place, row_place = row + init; curr_place < (place + ((bmh.w + 15) / 16) * 2); curr_place++ )
												{
													for( curr_pix = 128; curr_pix; curr_pix = curr_pix >> 1 )
													{
														/* check if theres a pixel in the bitplane and or it in the
														   chunky pixel */
														if( (curr_pix & *curr_place) && (!bmh.masking || (curr_plane != (bmh.nPlanes - 1))) )
															*row_place |= plane_bit;
														row_place += skip;
													}
												}
												place = curr_place;
											}
											else
											{
												/* its compressed so we need to unpack, its just a simple
												   RLE packer so this is simple */
												int n, count = ((bmh.w + 15) / 16) * 2;
												char *curr_place, *row_place;
												
												row_place = row + init;
												curr_place = place;
												while( count > 0 )
												{
													n = *curr_place;
													curr_place++;
													if( n >= 0 )
													{
														n++;
														count -= n;
														if( count < 0 ){ dead = true; break; }
														for( ; n; n--, curr_place++ )
														{
															for( curr_pix = 128; curr_pix; curr_pix = curr_pix >> 1 )
															{
																if( (curr_pix & *curr_place) && (!bmh.masking || (curr_plane != (bmh.nPlanes - 1))) )
																{
																	*row_place |= plane_bit;
																}
																row_place += skip;
															}
														}
													}
													else if( n != (-128) )
													{
														n = (-n) + 1;
														count -= n;
														if( count < 0 ){ dead = true; break; }
														for( ; n; n-- )
														{
															for( curr_pix = 128; curr_pix; curr_pix = curr_pix >> 1 )
															{
																if( curr_pix & *curr_place && (!bmh.masking || (curr_plane != (bmh.nPlanes - 1))) )
																{
																	*row_place |= plane_bit;
																}
																row_place += skip;
															}
														}
														curr_place++;
													}
												}
												place = curr_place;
											}
										}
										/* if its 8 bit we need to remap to the system cmap */
										if( hdr.colors == B_COLOR_8_BIT )
										{
											for( lpc = 0; lpc < hdr.rowBytes; lpc++ )
											{
												row[lpc] = real_cmap[(unsigned char)row[lpc]];
											}
											size = hdr.rowBytes;
										}
										else
											size = bmh.w * 4;
										if( ham )
										{
											/* its a ham so we need to convert from the 6 bit values
											   to the 24 bit values */
											struct ColorRegister *cr = (struct ColorRegister *)cmap;
											char *row_place, *row24_place;
											
											row_place = row;
											row24_place = row24;
											for( lpc = 0; lpc < rowWords; lpc++ )
											{
												switch( *row_place & 0x30 )
												{
													case 0x00:
														*row24_place = cr[*row_place].blue;
														row24_place++;
														*row24_place = cr[*row_place].green;
														row24_place++;
														*row24_place = cr[*row_place].red;
														row24_place++;
														row24_place++;
														row_place++;
														break;
													case 0x10:
														*row24_place = (*row_place & 0x0f) | ((*row_place & 0x0f) << 4);
														row24_place++;
														*row24_place = *(row24_place - 4);
														row24_place++;
														*row24_place = *(row24_place - 4);
														row24_place++;
														row24_place++;
														row_place++;
														break;
													case 0x20:
														*row24_place = *(row24_place - 4);
														row24_place++;
														*row24_place = *(row24_place - 4);
														row24_place++;
														*row24_place = (*row_place & 0x0f) | ((*row_place & 0x0f) << 4);
														row24_place++;
														row24_place++;
														row_place++;
														break;
													case 0x30:
														*row24_place = *(row24_place - 4);
														row24_place++;
														*row24_place = (*row_place & 0x0f) | ((*row_place & 0x0f) << 4);
														row24_place++;
														*row24_place = *(row24_place - 4);
														row24_place++;
														row24_place++;
														row_place++;
														break;
												}
											}
										}
										if( ham )
										{
											outStream->Write( row24, size );
										}
										else
										{
											outStream->Write( row, size );
										}
									}
								}
								else
									dead = true;
								if( body )
									free( body );
								if( row )
									free( row );
								if( ham && row24 )
								{
									row24 -= 4;
									free( row24 );
								}
							}
							break;
						default:
							{
								char chunk_name[5];
							
								chunk_name[0] = ((char *)&curr_chunk)[0];
								chunk_name[1] = ((char *)&curr_chunk)[1];
								chunk_name[2] = ((char *)&curr_chunk)[2];
								chunk_name[3] = ((char *)&curr_chunk)[3];
								chunk_name[4] = 0;
								printf( "Chunk: %s not recognized\n", chunk_name );
								inStream.Seek( inStream.Position() + chunk_size );
							}
							break;
					}
				}
			}
		}
	}
	if( cmap )
		free( cmap );
	if( real_cmap )
		free( real_cmap );
	return( isDataBitmap && !dead );
}

long
Translate(
	DStream &			inStream,
	const DATAInfo &	inInfo,
	BMessage *			/* ioExtension */,
	ulong				outFormat,
	DStream &			outStream)
{
	/*	Check that we handle input and output types */

	if( inInfo.formatType != DATA_BITMAP )
		return( DATA_NO_HANDLER );

	if( outFormat == 0 )
		outFormat = DATA_BITMAP;
	
	if( outFormat != DATA_BITMAP )
		return( DATA_NO_HANDLER );

	if( process_chunks( inStream, &outStream, MODE_TRANSLATE ) )
		return B_OK;
	else
		return( DATA_NO_HANDLER );
}


